with Messages, Ada.Integer_Text_IO;
use Messages, Ada.Integer_Text_IO;

package body Combined_Fourier_Ops is

  ScalZero: constant Scalar := Scal(Zero);
  ScalOne: constant Scalar := Scal(One);
  NumZero: constant Numeric := Numeric(Zero);
  NumOne: constant Numeric := Numeric(One);
  Trunc: constant Boolean := IsNumeric(ScalZero);

  procedure PutNumeric(Hs: in S_Fourier; Hn: out N_Fourier) is
    use Scalar_Part;
    S: Scalar;
  begin
    Numeric_Part.SetZero(Par(Hs),Rho(Hs),Hn);
    for M in Freq1 loop
      for N in Freq2 loop
        for K in Index3 loop
          S := Component(M,N,K,Hs);
          if S /= Scal(0) then
            Numeric_Part.SetComponent(M,N,K,Numeric(Center(S)),Hn);
          end if;
        end loop;
      end loop;
    end loop;
  end PutNumeric;

  function PutNumeric(Hs: S_Fourier) return N_Fourier is
    Hn: N_Fourier;
  begin
    PutNumeric(Hs,Hn);
    return Hn;
  end PutNumeric;

  procedure GetNumeric(Hn: in N_Fourier; Hs: out S_Fourier) is
    use Numeric_Part;
    S: Numeric;
  begin
    Scalar_Part.SetZero(Par(Hn),Rho(Hn),Hs);
    for M in Freq1 loop
      for N in Freq2 loop
        for K in Index3 loop
          S := Component(M,N,K,Hn);
          if S /= Numeric(0) then
            Scalar_Part.SetComponent(M,N,K,Scal(Rep(S)),Hs);
          end if;
        end loop;
      end loop;
    end loop;
  end GetNumeric;

  function GetNumeric(Hn: N_Fourier) return S_Fourier is
    Hs: S_Fourier;
  begin
    GetNumeric(Hn,Hs);
    return Hs;
  end GetNumeric;

  procedure Inv(Err: in Radius; R: in Weights; H1: in S_Fourier; H2: out S_Fourier) is
    use Scalar_Part;
    Ea,Eb,Ec,Ed,Er: Error;
    Hn1,Hn2: N_Fourier;
  begin
    PutNumeric(H1,Hn1);
    Numeric_Part.NumInv(Err,Hn1,Hn2);
    GetNumeric(Hn2,H2);
    if Trunc then
      SetRho(R,H2);
      return;
    end if;
    declare
      Ht: S_Fourier;
    begin
      Prod(R,H1,H2,Ht);
      AddComponent(0,0,0,-ScalOne,Ht);
      Ea := Norm(Ht);
      Eb := Norm(R,H1);
      Ec := Norm(R,H2)*Ea;
      Ed := Two*Ea+Two*Eb*Ec;
    end;
    if Sup(Ed)>Half or Sup(Ed+Two*Eb*Ec)>=One then
---      Message("Combined_Fourier_Ops.Inv",Domain_Violation);           --- continue
      Message("Combined_Fourier_Ops.Inv",Domain_Error);           --- stop
      Put("  Sup(Ea) = "); Put(Current_Output,Sup(Ea),5,3); New_Line;
      Put("  Sup(Eb) = "); Put(Current_Output,Sup(Eb),5,3); New_Line;
      Put("  Sup(Ec) = "); Put(Current_Output,Sup(Ec),5,3); New_Line;
      Put("  Sup(Ed) = "); Put(Current_Output,Sup(Ed),5,3); New_Line;
    end if;
    Er := Ec*Inv(ScalOne-Scal(Ed));
    AddComponent(0,0,0,Scal(Er),H2);
  end Inv;

  function Inv(Err: Radius; R: Weights; H: S_Fourier) return S_Fourier is
    Hr: S_Fourier;
  begin
    Inv(Err,R,H,Hr);
    return Hr;
  end Inv;

  procedure FixComp3(Err: in Radius; Rp3: in Radius; H: in S_Fourier; Hz: out S_Fourier) is
    use Scalar_Part;
 -- Rp3 is desired z-radius for Hz, Rho(Hz)(3) may be chosen smaller
    Repeat: Integer;
    R3: Radius;
    R: Weights;
    Ez,Ea,Eb,Ec,Er: Error;
    Sa1: Scalar;
    H3: S_Fourier;
  begin
    Hz := GetNumeric(Numeric_Part.NumFixComp3(Err,PutNumeric(H)));
    R := Rho(H);
    R3 := Rp3;
    Repeat := 2;
    while Repeat>0 loop                                 ---- allow decreasing z-domain
      R(3) := R3;
      SetRho(R,Hz);
      if Trunc then
        return;
      end if;
      Copy(Hz,H3);
      AddComponent(0,0,1,ScalOne,H3);                   ---- H3(.,.,z)=z+Hz(.,.,z)
      Ec := Norm(R,Comp3(H,H3)+Hz);                     ---- -H(.,.,H3) should be approx Hz
      Ez := Norm(Hz);
      R(3) := Sup(Norm(H3));
      Ea := Norm(R,Der3(R(3),H));
      Sa1 := ScalOne-Scal(Ea);
      Er := Two*Ec/Sa1;                                 ---- checks also that a<1
      AddComponent(0,0,0,Scal(Er),H3);
      R(3) := Sup(Norm(H3));
      if R(3)<Rho(H)(3) then
        Repeat := 0;
      else
        Repeat := Repeat-1;
        Put_Line("Combined_Fourier_Ops.FixComp3, Warning: Shrinking domain");
        Put("Rho(H)(3) = "); Put(Current_Output,Rho(H)(3),5,3); New_Line;
        Put("   ||H3|| = "); Put(Current_Output,R(3),5,3); New_Line;
        Put("old    R3 = "); Put(Current_Output,R3,5,3); New_Line;
        R3 := Rho(H)(3)-Sup(Ez+Er);
        Put("new    R3 = "); Put(Current_Output,R3,5,3); New_Line;
        if Repeat=0 then
          Message("Combined_Fourier_Ops.FixComp3",Giving_Up);
        end if;
      end if;
    end loop;
    Eb := Norm(R,DerDer3(R(3),H));
    if Sup(Four*Eb*Ec)>Inf(Sa1*Sa1) then
 --   Message("Combined_Fourier_Ops.FixComp3",Domain_Violation);       --- continue
      Message("Combined_Fourier_Ops.FixComp3",Domain_Error);           --- stop
      Put("  Sup(Ea) = "); Put(Current_Output,Sup(Ea),5,3); New_Line;
      Put("  Sup(Eb) = "); Put(Current_Output,Sup(Eb),5,3); New_Line;
      Put("  Sup(Ec) = "); Put(Current_Output,Sup(Ec),5,3); New_Line;
      Put("  Sup(Er) = "); Put(Current_Output,Sup(Er),5,3); New_Line;
    end if;
    H3 := Comp3(Der3(R(3),H),H3);                     ---- now H3 contains A(xi)
    AddComponent(0,0,0,-ScalOne,H3);
    R(3):=R3;                                         ---- reset to domain of Hz
    AddComponent(0,0,0,Scal(Ec*Norm(Inv(Err,R,H3))),Hz);
  end FixComp3;

end Combined_Fourier_Ops;




