with Ada.Text_IO, Ada.Integer_Text_IO, Rep_Ops;
use Ada.Text_IO, Ada.Integer_Text_IO, Rep_Ops;
with Ada.Sequential_IO;

package Linear_Ops is

   type Trip is
      record
         I: Integer;
         J: Integer;
         R: Rep;
      end record;

   type Vector is array(Integer range <>) of Rep;
   type Pivot is array(Integer range <>) of Integer;
   type Matrix is array(Integer range <>, Integer range <>) of Rep;
   type Sparse is array(Integer range <>) of Trip;                            -- range needs to start at zero

   Matrix_Singular: exception;
   Getting_Tired: exception;
   Matrix_Full: exception;

   package Sequential_Rep_IO is new Ada.Sequential_IO(Rep);
   package Sequential_Trip_IO is new Ada.Sequential_IO(Trip);

   procedure Read(Name: in String; B: out Vector; Bin: in Boolean := False);  -- Read B from Name (text/binary)
   procedure Read(Name: in String; A: out Matrix; Bin: in Boolean := False);  -- Read A from Name (text/binary)
   procedure Read(Name: in String; W: out Sparse; Bin: in Boolean := False);  -- Read W from Name (text/binary)
   procedure Write(Name: in String; B: in Vector; Bin: in Boolean := False);  -- Write B to Name (text/binary)
   procedure Write(Name: in String; A: in Matrix; Bin: in Boolean := False);  -- Write A to Name (text/binary)
   procedure Write(Name: in String; W: in Sparse; Bin: in Boolean := False);  -- Write W to Name (text/binary)

   procedure Show(P: in Pivot; L: in Integer := Integer'Last);                -- Display P(I) with I<=L
   procedure Show(B: in Vector; Cut: in Rep := Zero);                         -- Display elements>Cut of B
   procedure Show(B1,B2: in Vector; Cut: in Rep := Zero);                     -- Display elements>Cut of B1 and B2
   procedure ShowAll(A: in Matrix);
   procedure Show(A: in Matrix; Cut: in Rep := Zero);                         -- Display elements>Cut of A
   procedure Show(A1,A2: in Matrix; Cut: in Rep := Zero);                     -- Display elements>Cut of A1 and A2
   procedure Show(W: in Sparse; Cut: in Rep := Zero);                         -- Display elements>Cut of W

   function Length(W: in Sparse) return Integer;                              -- Number of filled (nonzero) elements
   procedure Component(I,J: in Integer; W: in out Sparse; R: out Rep);        -- R := W(i,j)
   procedure AddComponent(U: in Trip; W : in out Sparse; Cut: in Rep := Zero);-- Add R to W(i,j) if >Cut
   procedure AddColumn(J: in Integer; V: in Vector; W: in out Sparse;
                                                    Cut: in Rep := Zero);     -- Add V(i) to W(i,j) if >Cut
   procedure Cleanup(W: in out Sparse; Force: in Boolean := False);           -- Sort, remove zeros, combine overlaps
   procedure RowMap(W: in out Sparse; P: out Pivot; L: out Integer);          -- sorted list of row indices
   procedure MapIndex(P1,P2: in Pivot; W: in out Sparse);                     -- Apply maps P1 and P2 to indices of W
   procedure InvMapIndex(P1,P2: in Pivot; L1,L2: in Integer;                  -- Inverse of MapIndex
                                          W: in out Sparse);                  --  assuming P1 and P2 are increasing

   function Norm1(B: Vector) return Rep;                                      -- Standard L1 norm of B
   function Norm1(Bf,B: Vector) return Rep;                                   -- L1 norm of B with weights Bf
   function Norm1(A: Matrix) return Rep;                                      -- Operator norm of A, for standard L1
   function Norm1(Bf: Vector; A: Matrix) return Rep;                          -- Operator norm of A, for weighted L1
   function Norm2(B: Vector) return Rep;                                      -- L2 norm of B
   function IProd(Bf,B1,B2: Vector) return Rep;                               -- L2 inner product with weight Bf

   procedure SetZero(B: out Vector);                                          -- B := 0
   procedure SetZero(A: out Matrix);                                          -- A := 0
   procedure SetZero(W: out Sparse);                                          -- W := 0
   procedure SetColumn(J: in Integer; V: in Vector; A: in out Matrix);        -- A(.,J) := V
   procedure Transpose(W: in out Sparse);                                     -- Transpose W
   procedure Copy(A: in Matrix; W: out Sparse; Cut: in Rep := Zero);          -- W := A  (plain)
   procedure Copy(A: in Matrix; P1,P2: in Pivot; W: out Sparse;
                                                 Cut: in Rep := Zero);        -- W := A  (re-indexed using P1,P2)
   procedure Copy(W: in Sparse; A: out Matrix; Check: in Boolean := True);    -- A := W  (plain)
   procedure Copy(W: in out Sparse; A: out Matrix; P: out Pivot;
                                                   L: out Integer);           -- A := W  (re-indexed using P inverse)
   procedure Copy(W: in out Sparse; A: out Matrix; P1,P2: out Pivot;
                                                   L1,L2: out Integer);       -- A := W  (re-indexed using P1,P2 inverse)

   procedure Add(B1: in Vector; B2: in out Vector);                           -- B2 := B2+B1
   procedure Add(A1: in Matrix; A2: in out Matrix);                           -- A2 := A2+A1
   procedure Add(R: in Rep; A: in out Matrix);                                -- A := A+R*Identity
   procedure Sub(B1: in Vector; B2: in out Vector);                           -- B2 := B2-B1
   procedure Sub(A1: in Matrix; A2: in out Matrix);                           -- A2 := A2-A1
   procedure Mult(R: in Rep; B: in out Vector);                               -- B := R*B
   procedure Mult(R: in Rep; A: in out Matrix);                               -- A := R*A
   procedure Mult(A1: in Matrix; A2: in out Matrix);                          -- A2 := A1*A2
   procedure DMult(Bf: in Vector; B: in out Vector);                          -- B := Diag(Bf)*B
   procedure InvDMult(Bf: in Vector; B: in out Vector);                       -- B := (Id/Diag(Bf))*B
   procedure Cutoff(Rf: in Rep; Bf,B1: in Vector; N1,N2: out Integer;
                                                  B2: out Vector);            -- B2 := B1 part >= Rf*Bf

   function "-"(B: Vector) return Vector;                                     -- -B
   function "+"(B1,B2: Vector) return Vector;                                 -- B1+B2
   function "+"(A1,A2: in Matrix) return Matrix;                              -- A1+A2
   function "-"(B1,B2: Vector) return Vector;                                 -- B1-B2
   function "-"(A1,A2: in Matrix) return Matrix;                              -- A1-A2
   function "*"(B1,B2: Vector) return Rep;                                    -- L2 inner product <B1,B2>
   function "*"(R: Rep; B: Vector) return Vector;                             -- R*B
   function "*"(R: Rep; A: Matrix) return Matrix;                             -- R*A
   function "*"(B1,B2: Vector) return Vector;                                 -- Diag(B1)*B2
   function "*"(A: Matrix; B: Vector) return Vector;                          -- A*B
   function "*"(B: Vector; A: Matrix) return Vector;                          -- Transpose(A)*B
   function "*"(A1,A2: Matrix) return Matrix;                                 -- A1*A2
   function "*"(W: Sparse; B: Vector) return Vector;                          -- W*B
   function "*"(B: Vector; W: Sparse) return Vector;                          -- Transpose(W)*B
   function PlusMinusDet(A: Matrix) return Rep;
   procedure Balanc(A: in out Matrix);                                        -- Balance A, preserves spectrum
   procedure Decomp(A: in out Matrix;
                    Cond: out Rep; Pvt: out Pivot; Work: out Vector);         -- Triangularize A, LU decomposition
   procedure Solve(A: in Matrix; B: in out Vector; Pvt: in Pivot);            -- Solve A*B2=B1 for triangularized A
   procedure Solve(W: in Sparse; B: in Vector; X: in out Vector;
                                               Rsq: out Rep;
                                               Eps: in Rep := 1.0E-5);        -- Solve W*X=B (steepest descent)
   procedure DiagFirst(W: in out Sparse);                                     -- Put diagonal elements at the beginning
   procedure NeuSolve(W: in Sparse; B: in Vector; X: in out Vector;
                                                  Eps: in Rep := 1.0E-8);     -- Solve W*X=B (Neumann serise) assuming DiagFirst
   procedure BCG(W: in Sparse; B: in Vector; X: in out Vector;
                                             Eps: in Rep := 1.0E-8);          -- Solve W*X=B (biconjugate gradient)
   procedure BCGstab(W: in Sparse; B: in Vector; X: in out Vector;
                                                 Eps: in Rep := 1.0E-8);      -- Solve W*X=B (biconjugate gradient stabilized)
   procedure BCGstab(W: in Sparse; Bf,B: in Vector; X: in out Vector;
                                                    Eabs: in Rep := 1.0E-8;
                                                    Erel: in Rep := 1.0E-1);  -- BCGstab with Bf-inner product

   procedure Invert(A: in out Matrix);                                        -- A := 1/A
   procedure Eigenvalues(A: in Matrix; Br,Bi: out Vector);                    -- Eigenvalues of A (real and imag parts)
   procedure Eigenvector(A: in Matrix; R: in Rep; B: out Vector;
                                                  Eps: in Rep := 1.0E-8);     -- Try to solve (A-R)*B=0

private

   type TripCompare is access function (U,V: Trip) return Integer;

   procedure ShortPut(F: in File_Type; I: in Integer);                        -- Write I in shortest format

   procedure BinRead(Name: in String; B: out Vector);                         -- Read B from Name, binary format
   procedure TextRead(Name: in String; B: out Vector);                        -- Read B from Name, ascii format
   procedure BinRead(Name: in String; A: out Matrix);                         -- Read A from Name, binary format
   procedure TextRead(Name: in String; A: out Matrix);                        -- Read A from Name, ascii format
   procedure BinRead(Name: in String; W: out Sparse);                         -- Read W from Name, binary format
   procedure TextRead(Name: in String; W: out Sparse);                        -- Read A from Name, ascii format
   procedure BinWrite(Name: in String; B: in Vector);                         -- Write B to Name, binary format
   procedure TextWrite(Name: in String; B: in Vector);                        -- Write B to Name, ascii format
   procedure BinWrite(Name: in String; A: in Matrix);                         -- Write A to Name, binary format
   procedure TextWrite(Name: in String; A: in Matrix);                        -- Write A to Name, ascii format
   procedure BinWrite(Name: in String; W: in Sparse);                         -- Write W to Name, binary format
   procedure TextWrite(Name: in String; W: in Sparse);                        -- Write W to Name, ascii format

   procedure ElmHes(A: in out Matrix);                                        -- Reduce A to Hessenberg form
   procedure Hqr(A: in out Matrix; Wr,Wi: out Vector);                        -- Find Eigenvalues of Hessenberg matrix
   function CompareIJ(U,V: Trip) return Integer;                              -- Compare .I and .J components of U,V
   function MatchIJ(U: Trip; W: Sparse; Kmax: Integer;
                                        Kmin: Integer := 1) return Integer;   -- Find K for which W(K)=(U.I,U.J,.)
   procedure Sort(W: in out Sparse; Compare: in TripCompare);                 -- Sort W
   procedure Merge(P1,P2: in Pivot; L1,L2: in Integer; P: out Pivot;
                                                       L: out Integer);       -- merge sorted lists (duplicates removed)

end Linear_Ops;




