with Messages, Ada.Integer_Text_IO;
use Messages, Ada.Integer_Text_IO;

package body Taylor_Ops is


   ScalZero: constant Scalar := Scal(Zero);
   ScalOne: constant Scalar := Scal(One);
   Trunc: constant Boolean := IsNumeric(ScalZero);


   function IMin(I1,I2: Integer) return Integer renames Integer'Min;
   function IMax(I1,I2: Integer) return Integer renames Integer'Max;

   function Deg(P: Taylor) return Degree is
   begin
      return P.D;
   end Deg;


   function EffDeg(P: Taylor) return Degree is
   begin
      for K in reverse 0 .. P.D loop
         if P.C(K) /= ScalZero then
            return K;
         end if;
      end loop;
      return -1;
   end EffDeg;


   function Component(K: Natural; P: Taylor) return Scalar is
   begin
      if K > P.D then
         return ScalZero;
      else
         return P.C(K);
      end if;
   end Component;


   procedure SetComponent(D: in Degree; S: in Scalar; P: in out Taylor) is
   begin
      if D > P.D then
         for K in P.D+1 .. D-1 loop
            P.C(K) := ScalZero;
         end loop;
         P.D := D;
      end if;
      P.C(D) := S;
   end SetComponent;


   procedure AddComponent(D: in Degree; S: in Scalar; P: in out Taylor) is
   begin
      if S /= ScalZero then
         if D > P.D then
            for K in P.D+1 .. D-1 loop
               P.C(K) := ScalZero;
            end loop;
            P.D := D;
            P.C(D) := S;
         else
            P.C(D) := P.C(D)+S;
         end if;
      end if;
   end AddComponent;


   procedure SetZero(P: out Taylor) is
   begin
      P.D := -1;
   end SetZero;


   procedure Get(F: in File_Type; P: out Taylor) is
      D: Integer;
      C: Scalar;
   begin
      Get(File=>F, Item=>D);
      P.D := IMin(D,Size);
      for K in 0 .. D loop
         Get(F,C);
         if K <= Size then
            P.C(K) := C;
         end if;
      end loop;
      P.D := EffDeg(P);
   end Get;


   procedure Put(F: in File_Type; P: in Taylor) is
   begin
      Put(File=>F, Item=>P.D, Width=>3);
      New_Line(F);
      for K in 0 .. P.D loop
         Put(F,Center(P.C(K)));
         New_Line(F);
      end loop;
   end Put;


   procedure Read(Name: in String; P: out Taylor) is
      F: File_Type;
   begin
      Open(F,In_File,Name);
      Get(F,P);
      Close(F);
   end Read;


   procedure Write(Name: in String; P: in Taylor) is
      F: File_Type;
   begin
      Create(F,Out_File,Name);
      Put(F,P);
      Close(F);
   end Write;


   procedure Show(P: in Taylor) is
   begin
      Put(P.D);
      New_Line;
      for K in 0 .. P.D loop
         Show(P.C(K));
      end loop;
      New_Line;
   end Show;


   procedure Cshow(F: in File_Type; P: in Taylor) is
   begin

      if (P.D>=0) then
         Put(F,"s=");  Put(F,Center(P.C(P.D))); Put(F,";");
         New_Line(F);
         for K in reverse 0 .. P.D-1 loop
            if (P.C(K)=ScalZero) then
               Put(F," s=z*s;");
               New_Line(F);
            else
               Put(F,"s=z*s+("); Put(F,Center(P.C(K)));Put(F,");");
               New_Line(F);
            end if;
         end loop;
      end if;
   end Cshow;


   function ErrCoeff(R: Radius; K: Natural; P: Taylor) return Error is
   begin
      if Trunc then
         return Zero_Error;
      end if;
      declare
         D: Integer;
         E: Error;
      begin
         E := Zero_Error;
         D := IMin(K,P.D);
         for N in 0 .. D loop
            E := E/R+ErrPart(P.C(N));
         end loop;
         return E*(Unit_Error/R)**(K-D);
      end;
   end ErrCoeff;


   function Coeff(R: Radius; K: Natural; P: Taylor) return Scalar is
   begin
      return ResetErr(Component(K,P))+Scal(ErrCoeff(R,K,P));
   end Coeff;


   function Val(R: Radius; Z: Scalar; P: Taylor) return Scalar is
      S: Scalar;
   begin
      if not Trunc and then SupAbs(Z) >= R then
         Message("Taylor_Ops.Val",Domain_Violation);
      end if;
      S := ScalZero;
      for K in reverse 0 .. P.D loop
         S := Z*S+P.C(K);
      end loop;
      return S;
   end Val;


   procedure MaxAbsComponent(P: Taylor; R: in out Rep) is
   begin
      for K in 0 .. P.D loop
         R := Max(R,SupAbs(P.C(K)));
      end loop;
   end MaxAbsComponent;


   function Norm(R: Radius; P: Taylor) return Error is
      E: Error;
   begin
      E := Zero_Error;
      if P.D/=-1  then
         for K in reverse 0 .. P.D loop
            if (Sup(E)>Rep(1.0E+150)) then
               Message("  Taylor_Ops.Norm.Blow_UP",Giving_Up);
            end if;
            E := R*E+Err(P.C(K));
         end loop;
      end if;
      return E;
   end Norm;


   function SupNorm(R: Radius; P: Taylor) return Error is
      E,Er: Error;
   begin
      Er := Zero_Error;
      if P.D/=-1  then
         for K in  0 .. P.D loop
            E :=Err(P.C(K));
            for I in 1..K loop
               E := R*E;
               if (Sup(E)>Rep(1.0E+250)) then
                  Message("  Taylor_Ops.SupNorm.Blow_UP",Giving_Up);
               end if;
            end loop;
            if Sup(E)>Sup(Er) then Er:=E; end if;
         end loop;
      end if;
      return Er;
   end SupNorm;


   procedure Copy(P1: in Taylor; P2: out Taylor) is
   begin
      P2.D := P1.D;
      for K in 0 .. P1.D loop
         P2.C(K) := P1.C(K);
      end loop;
   end Copy;


   procedure OddEliminate(P1: in Taylor; P2: out Taylor) is
   begin
      SetZero(P2);
      P2.D := P1.D;
      for K in 0 .. P1.D loop
         if ((K mod 2)=0) then
            P2.C(K) := P1.C(K);
         else
            P2.C(K):=ScalZero;
         end if;
      end loop;
   end OddEliminate;


   procedure EvenEliminate(P1: in Taylor; P2: out Taylor) is
   begin
      SetZero(P2);
      P2.D := P1.D;
      for K in 0 .. P1.D loop
         if ((K mod 2=1)) then
            P2.C(K) := P1.C(K);
         else
            P2.C(K):=ScalZero;
         end if;
      end loop;
   end EvenEliminate;


   procedure CopyAbs(P1: in Taylor; P2: out Taylor) is
   begin
      P2.D := P1.D;
      for K in 0 .. P1.D loop
         P2.C(K) := abs(P1.C(K));
      end loop;
   end CopyAbs;


   procedure CopySign(P1: in Taylor; P2: in out Taylor) is
   begin
      for K in 0 .. P1.D loop

         if (Center(P1.C(K))>Zero) then P2.C(K) := abs(P2.C(K));
         else P2.C(K) := Scal(-1.0)*abs(P2.C(K));

         end if;
      end loop;
   end CopySign;


   procedure Neg(P1: in Taylor; P2: out Taylor) is
   begin
      P2.D := P1.D;
      for K in 0 .. P1.D loop
         P2.C(K) := -P1.C(K);
      end loop;
   end Neg;


   function "-"(P: Taylor) return Taylor is
      Pt: Taylor;
   begin
      Neg(P,Pt);
      return Pt;
   end "-";


   procedure Mult(S: in Scalar; P: in out Taylor) is
   begin
      for K in 0 .. P.D loop
         P.C(K) := S*P.C(K);
      end loop;
   end Mult;


   procedure Prod(S: in Scalar; P1: in Taylor; P2: out Taylor) is
   begin
      if S = ScalZero then
         P2.D := -1;
      else
         P2.D := P1.D;
         for O in 0 .. P1.D loop
            P2.C(O) := S*P1.C(O);
         end loop;
      end if;
   end Prod;


   function "*"(S: Scalar; P: Taylor) return Taylor is
      Pt: Taylor;
   begin
      Prod(S,P,Pt);
      return Pt;
   end "*";


   procedure Add(P1: in Taylor; P2: in out Taylor) is
   begin
      for K in 0 .. IMin(P1.D,P2.D) loop
         P2.C(K) := P2.C(K)+P1.C(K);
      end loop;
      if P2.D < P1.D then
         for K in P2.D+1 .. P1.D loop
            P2.C(K) := P1.C(K);
         end loop;
         P2.D := P1.D;
      end if;
   end Add;


   procedure Sum(P1,P2: in Taylor; P3: out Taylor) is
   begin
      P3.D := IMax(P1.D,P2.D);
      for K in 0 .. IMin(P1.D,P2.D) loop
         P3.C(K) := P1.C(K)+P2.C(K);
      end loop;
      for K in P2.D+1 .. P3.D loop
         P3.C(K) := P1.C(K);
      end loop;
      for K in P1.D+1 .. P3.D loop
         P3.C(K) := P2.C(K);
      end loop;
   end Sum;


   function "+"(P1,P2: Taylor) return Taylor is
      Pt: Taylor;
   begin
      Sum(P1,P2,Pt);
      return Pt;
   end "+";


   procedure Sub(P1: in Taylor; P2: in out Taylor) is
   begin
      for K in 0 .. IMin(P1.D,P2.D) loop
         P2.C(K) := P2.C(K)-P1.C(K);
      end loop;
      if P2.D < P1.D then
         for K in P2.D+1 .. P1.D loop
            P2.C(K) := -P1.C(K);
         end loop;
         P2.D := P1.D;
      end if;
   end Sub;


   procedure Diff(P1,P2: in Taylor; P3: out Taylor) is
   begin
      P3.D := IMax(P1.D,P2.D);
      for K in 0 .. IMin(P1.D,P2.D) loop
         P3.C(K) := P1.C(K)-P2.C(K);
      end loop;
      for K in P2.D+1 .. P3.D loop
         P3.C(K) := P1.C(K);
      end loop;
      for K in P1.D+1 .. P3.D loop
         P3.C(K) := -P2.C(K);
      end loop;
   end Diff;


   function "-"(P1,P2: Taylor) return Taylor is
      Pt: Taylor;
   begin
      Diff(P1,P2,Pt);
      return Pt;
   end "-";


   procedure MultAdd(S: in Scalar; P1: in Taylor; P2: in out Taylor) is
   begin
      if S /= ScalZero then
         for K in 0 .. IMin(P1.D,P2.D) loop
            P2.C(K) := P2.C(K)+S*P1.C(K);
         end loop;
         if P2.D < P1.D then
            for K in P2.D+1 .. P1.D loop
               P2.C(K) := S*P1.C(K);
            end loop;
            P2.D := P1.D;
         end if;
      end if;
   end Multadd;


   procedure Scale(Z,S: in Scalar; P1: in Taylor; P2: out Taylor) is
      Sp: Scalar;
   begin
      P2.D := P1.D;
      Sp := Z;
      for I in 0 .. P2.D loop
         P2.C(I) := Sp*P1.C(I);
         Sp := Sp*S;
      end loop;
   end Scale;


   function Scale(Z,S: Scalar; P: Taylor) return Taylor is
      Pt: Taylor;
   begin
      Scale(Z,S,P,Pt);
      return Pt;
   end Scale;

   procedure ScalePlus(L: in Cutoff; Z,S: in Scalar; R1,R2: in Radius; P1: in Taylor; P2: out Taylor) is
      Kmax: Integer;
      Sp,Sf: Scalar;
   begin
      Scale(Z,S,P1,P2);
      if not Trunc then
         Sf := S*Scal(R2)/Scal(R1);
         if Sup(Sf)>One then
            Message("Taylor_Ops.IplusScale",Domain_Violation);
         end if;
         Kmax := Imin(L,P2.D);
         Sp := Sf**(L-Kmax+1);
         for K in reverse 0 .. Kmax loop
            P2.C(K) := Sp*P2.C(K);                          -- use that errors are of degree > L
            Sp := Sf*Sp;
         end loop;
      end if;
   end ScalePlus;


   procedure Fmax1(R: in Scalar; M: out Natural; F: out Error) is
      X: Scalar;
   begin
      X := Inv(-Log(R));
      M := Integer(Rep'Ceiling(Sup(X)));
      F := Err(X*Exp(-ScalOne));
   end Fmax1;


   procedure Der(R1,R2: in Radius; P1: in Taylor; P2: out Taylor) is
   begin
      if P1.D >= 0 then
         for K in 1 .. P1.D loop
            P2.C(K-1) := Scal(K)*P1.C(K);
         end loop;
         P2.D := P1.D-1;
         if Trunc or else Err0(P1) then
            return;
         end if;
         if R2>=R1 then
            Message("Taylor_Ops.Der",Domain_Violation);
            Put("  R1 ="); Put(Current_Output,R1,3,3);
            Put("  R2 ="); Put(Current_Output,R2,3,3); New_Line;
         end if;
         declare
            M: Natural;
            E,F,Ri: Error;
         begin
            if P1.D=0 then
               P2.D := 0;
               P2.C(0) := ScalZero;
            end if;
            Ri := Err(R1)/R2;
            Fmax1(Scal(R2)/Scal(R1),M,F);
            E := ErrPart(P1.C(0))*(F/R2);
            for K in 1 .. IMin(M,P1.D) loop
               F := F*Ri;
               ErrMult(F/K,P2.C(K-1));
            end loop;
            P2.C(0) := P2.C(0)+Scal(E);
         end;
      else
         P2.D := -1;
      end if;
   end Der;


   function Der(R1,R2: Radius; P: Taylor) return Taylor is
      Pt: Taylor;
   begin
      Der(R1,R2,P,Pt);
      return Pt;
   end Der;


   procedure Fmax2(R: in Scalar; M: out Natural; F: out Error) is
      LogR,X: Scalar;
   begin
      LogR := Log(R);
      X := Inv(-LogR);
      X := X+Scal(Half)+Sqrt(X*X+Scal(Half*Half));
      M := Integer(Rep'Ceiling(Sup(X)));
      F := Err(X*(X-ScalOne)*Exp(X*LogR));
   end Fmax2;


   procedure DerDer(R1,R2: in Radius; P1: in Taylor; P2: out Taylor) is
   begin
      if P1.D >= 0 then
         for K in 2 .. P1.D loop
            P2.C(K-2) := Scal(K*(K-1))*P1.C(K);
         end loop;
         P2.D := Imax(P1.D-2,-1);
         if Trunc or else Err0(P1) then
            return;
         end if;
         if R2>=R1 then
            Message("Taylor_Ops.DerDer",Domain_Violation);
            Put("  R1 ="); Put(Current_Output,R1,3,3);
            Put("  R2 ="); Put(Current_Output,R2,3,3); New_Line;
         end if;
         declare
            M: Natural;
            E,F,Ri: Error;
         begin
            if P1.D<2 then
               P2.D := 0;
               P2.C(0) := ScalZero;
            end if;
            Ri := Err(R1)/R2;
            Fmax2(Scal(R2)/Scal(R1),M,F);
            E :=(ErrPart(Component(0,P1))+R1*ErrPart(Component(1,P1)))*((F/R2)/R2);
            for K in 2 .. IMin(M,P1.D) loop
               F := F*Ri;
               ErrMult(F/(K*(K-1)),P2.C(K-2));
            end loop;
            P2.C(0) := P2.C(0)+Scal(E);
         end;
      else
         P2.D := -1;
      end if;
   end DerDer;


   function DerDer(R1,R2: Radius; P: Taylor) return Taylor is
      Pt: Taylor;
   begin
      DerDer(R1,R2,P,Pt);
      return Pt;
   end DerDer;


   procedure DerMultPart(S: in Scalar; P: in out Taylor) is
   begin
      if P.D >= 0 then
         for K in 1 .. P.D loop
            P.C(K-1) := S*Scal(K)*P.C(K);
         end loop;
         P.D := P.D-1;
      end if;
   end DerMultPart;


   procedure DerProdPart(S: in Scalar; P1: in Taylor; P2: out Taylor) is
   begin
      if P1.D > 0 then
         for K in 1 .. P1.D loop
            P2.C(K-1) := S*Scal(K)*P1.C(K);
         end loop;
         P2.D := P1.D-1;
      else
         P2.D := -1;
      end if;
   end DerProdPart;


   function Center0(P: Taylor) return Boolean is
   begin
      if P.D<0 then
         return True;
      else
         for K in reverse 1 .. P.D loop
            if not Center0(P.C(K)) then
               return False;
            end if;
         end loop;
         return Center0(P.C(0));
      end if;
   end Center0;


   procedure ResetCenter(P: in out Taylor; Kmin: in Natural := 0; Kmax: in Cutoff := Size) is
   begin
      if P.D>=0 and Kmin<=Kmax and Kmin<=P.D then
         if Trunc then
            if Kmax<P.D then
               for K in Kmin .. Kmax loop
                  P.C(K) := ScalZero;
               end loop;
            else
               P.D := Kmin-1;
            end if;
         else
            for K in Kmin .. Imin(Kmax,P.D) loop
               ResetCenter(P.C(K));
            end loop;
         end if;
      end if;
   end ResetCenter;


   procedure ResetCenter(P: in Taylor; Pr: out Taylor; Kmin: in Natural := 0; Kmax: in Cutoff := Size) is
   begin
      if P.D>=0 and Kmin<=Kmax and Kmin<=P.D then
         for K in 0 .. Kmin-1 loop
            Pr.C(K) := P.C(K);
         end loop;
         if Kmax<P.D then
            for K in Kmin .. Kmax loop
               Pr.C(K) := ResetCenter(P.C(K));
            end loop;
            for K in Kmax+1 .. P.D loop
               Pr.C(K) := P.C(K);
            end loop;
            Pr.D := P.D;
         elsif Trunc then
            Pr.D := Kmin-1;
         else
            for K in Kmin .. P.D loop
               Pr.C(K) := ResetCenter(P.C(K));
            end loop;
            Pr.D := P.D;
         end if;
      else
         Copy(P,Pr);
      end if;
   end ResetCenter;


   function ResetCenter(P: Taylor; Kmin: Natural := 0; Kmax: Cutoff := Size) return Taylor is
      Pr: Taylor;
   begin
      ResetCenter(P,Pr,Kmin,Kmax);
      return Pr;
   end ResetCenter;


   function Err0(P: Taylor) return Boolean is
   begin
      if P.D<0 then
         return True;
      else
         for K in reverse 1 .. P.D loop
            if not Err0(P.C(K)) then
               return False;
            end if;
         end loop;
         return Err0(P.C(0));
      end if;
   end Err0;


   procedure ResetErr(P: in out Taylor; Kmin: in Natural := 0; Kmax: in Cutoff := Size) is
   begin
      if not Trunc then
         for K in Kmin .. Imin(Kmax,P.D) loop
            ResetErr(P.C(K));
         end loop;
      end if;
   end ResetErr;


   procedure ResetErr(P: in Taylor; Pr: out Taylor; Kmin: in Natural := 0; Kmax: in Cutoff := Size) is
   begin
      if not Trunc and then (P.D>=0 and Kmin<=Kmax and Kmin<=P.D) then
         for K in 0 .. Kmin-1 loop
            Pr.C(K) := P.C(K);
         end loop;
         for K in Kmin .. Imin(Kmax,P.D) loop
            Pr.C(K) := ResetErr(P.C(K));
         end loop;
         for K in Kmax+1 .. P.D loop
            Pr.C(K) := P.C(K);
         end loop;
         Pr.D := P.D;
      else
         Copy(P,Pr);
      end if;
   end ResetErr;


   function ResetErr(P: Taylor; Kmin: in Natural := 0; Kmax: in Cutoff := Size) return Taylor is
      Pr: Taylor;
   begin
      ResetErr(P,Pr,Kmin,Kmax);
      return Pr;
   end ResetErr;


   procedure Split(P: in Taylor; Pc,Pe: out Taylor) is
   begin
      Pc.D := P.D;
      if Trunc then
         Pe.D := -1;
         for K in 0 .. P.D loop
            Pc.C(K) := P.C(K);
         end loop;
      else
         Pe.D := P.D;
         for K in 0 .. P.D loop
            Split(P.C(K),Pc.C(K),Pe.C(K));
         end loop;
      end if;
   end Split;


   procedure ErrConvert(P: in Taylor; Pe: out Taylor) is
   begin
      if Trunc then
         Pe.D := -1;
      else
         Pe.D := P.D;
         for K in 0 .. P.D loop
            Pe.C(K) := Scal(Err(P.C(K)));
         end loop;
      end if;
   end ErrConvert;


   function ErrConvert(P: Taylor) return Taylor is
      Pe: Taylor;
   begin
      ErrConvert(P,Pe);
      return Pe;
   end ErrConvert;


   procedure DegConvert(R: in Radius; D: in Degree; P: in out Taylor) is
   begin
      if D<P.D then
         if Trunc then
            P.D := D;
         elsif D<0 then
            Message("Taylor_Ops.DegConvert",Index_Error);
         else
            declare
               E: Error;
            begin
               E := Err(P.C(P.D));
               for K in reverse D+1 .. P.D-1 loop
                  E := R*E+Err(P.C(K));
               end loop;
               P.C(D) := Scal(R*E)+P.C(D);
               P.D := D;
            end;
         end if;
      end if;
   end DegConvert;


   procedure DegConvert(R: in Radius; D: in Degree; P: in Taylor; Pc: out Taylor) is
   begin
      if P.D<0 then
         Pc.D := -1;
      elsif D>=0 then
         Pc.D := IMin(D,P.D);
         for K in 0 .. Pc.D loop
            Pc.C(K) := P.C(K);
         end loop;
         if not Trunc and then D<P.D then
            declare
               E: Error;
            begin
               E := Err(P.C(P.D));
               for K in reverse D+1 .. P.D-1 loop
                  E := R*E+Err(P.C(K));
               end loop;
               Pc.C(D) := Scal(R*E)+Pc.C(D);
            end;
         end if;
      elsif Trunc then
         Pc.D := -1;
      else
         Message("Taylor_Ops.DegConvert",Index_Error);
      end if;
   end DegConvert;


   function DegConvert(R: Radius; D: Degree; P: Taylor) return Taylor is
      Pc: Taylor;
   begin
      DegConvert(R,D,P,Pc);
      return Pc;
   end DegConvert;


   procedure XMultAdd(R: in Radius; K: in Degree; S: in Scalar;
                                                  P1: in Taylor; P2: in out Taylor; D: in Degree := Size) is
      D1: Natural;
      D2: Degree;
   begin
      DegConvert(R,D,P2);                       ---- now P2.D <= D
      if S /= ScalZero and then P1.D >= 0 then
         D1 := K+P1.D;
         D2 := IMin(D1,D);
         if P2.D >= D2 then                      ---- l.o. stuff fits below P2.D
            for I in K .. D2 loop
               P2.C(I) := P2.C(I)+S*P1.C(I-K);
            end loop;
         elsif P2.D >= K then                    ---- l.o. stuff above and below P2.D
            for I in K .. P2.D loop
               P2.C(I) := P2.C(I)+S*P1.C(I-K);
            end loop;
            for I in P2.D+1 .. D2 loop
               P2.C(I) := S*P1.C(I-K);
            end loop;
            P2.D := D2;
         else                                    ---- no l.o. stuff below P2.D
            for I in P2.D+1 .. IMin(K-1,D2) loop
               P2.C(I) := ScalZero;
            end loop;
            for I in K .. D2 loop
               P2.C(I) := S*P1.C(I-K);
            end loop;
            P2.D := D2;
         end if;
         if not Trunc and then D1>D then         ---- add h.o. stuff
            declare
               E: Error;
            begin
               D2 := IMax(K,D+1);
               E := Zero_Error;
               for I in reverse D2 .. D1 loop
                  E := R*E+Err(S*P1.C(I-K));
               end loop;
               P2.C(D) := Scal((Err(R)**(D2-D))*E)+P2.C(D);
            end;
         end if;
      end if;
   end XMultAdd;


   procedure Mult(R: in Radius; P1: in Taylor; P2: in out Taylor; D: in Degree := Size) is
      D1: Natural;
      D2: Degree;
      S: Scalar;
   begin
      if P2.D >= 0 then
         if P1.D >= 0 then
            D1 := P1.D+P2.D;
            S := ScalZero;
            if not Trunc then
               declare
                  E: Scalar;
               begin
                  for K in reverse D+1 .. D1 loop                ---- collect h.o. terms
                     E := ScalZero;
                     for I in IMax(0,K-P1.D) .. IMin(K,P2.D) loop
                        E := E+P2.C(I)*P1.C(K-I);
                     end loop;
                     S := Scal(R)*S+E;
                  end loop;
                  S := Scal(R*Err(S));
               end;
            end if;
            D2 := IMin(D1,D);
            for K in reverse 0 .. D2 loop                     ---- compute l.o. terms
               for I in IMax(0,K-P1.D) .. IMin(K,P2.D) loop
                  S := S+P2.C(I)*P1.C(K-I);
               end loop;
               P2.C(K) := S;
               S := ScalZero;
            end loop;
            P2.D := D2;
         else
            P2.D := -1;
         end if;
      end if;
   end Mult;


   procedure Prod(R: in Radius; P1,P2: in Taylor; P3: out Taylor; D: in Degree := Size) is
      D1: Natural;
      D2: Degree;
      S: Scalar;
   begin
      if P1.D >= 0 and P2.D >= 0 then
         D1 := P1.D+P2.D;
         S := ScalZero;
         if not Trunc then
            declare
               E: Scalar;
            begin
               for K in reverse D+1 .. D1 loop                ---- collect h.o. terms
                  E := ScalZero;
                  for I in IMax(0,K-P1.D) .. IMin(K,P2.D) loop
                     E := E+P2.C(I)*P1.C(K-I);
                  end loop;
                  S := Scal(R)*S+E;
               end loop;
               S := Scal(R*Err(S));
            end;
         end if;
         D2 := IMin(D1,D);
         for K in reverse 0 .. D2 loop                     ---- compute l.o. terms
            for I in IMax(0,K-P1.D) .. IMin(K,P2.D) loop
               S := S+P2.C(I)*P1.C(K-I);
            end loop;
            P3.C(K) := S;
            S := ScalZero;
         end loop;
         P3.D := D2;
      else
         P3.D := -1;
      end if;
   end Prod;


   function Prod(R: in Radius; P1,P2: Taylor; D: Degree := Size) return Taylor is
      Pt: Taylor;
   begin
      Prod(R,P1,P2,Pt,D);
      return Pt;
   end Prod;


   procedure MultAdd(R: in Radius; P1,P2: in Taylor; P3: in out Taylor; D: Degree := Size) is
      D1: Natural;
      D2: Degree;
   begin
      if abs(P3.D)>1000 then New_Line; Put("Wrong P3.D!!!!"); end if;
      if abs(P2.D)>1000 then New_Line; Put("Wrong P2.D!!!!"); end if;
      if abs(P1.D)>1000 then New_Line; Put("Wrong P1.D!!!!"); end if;
      DegConvert(R,D,P3);                                 ---- now P3.D <= D
      if P1.D >= 0 and P2.D >= 0 then
         D1 := P1.D+P2.D;
         D2 := IMin(D1,D);
         if P3.D<D2 then                                   ---- need to fill in 0s
            for I in P3.D+1 .. D2 loop
               P3.C(I) := ScalZero;
            end loop;
            P3.D := D2;
         end if;
         if not Trunc and then D1>D then                   ---- collect h.o. terms
            declare
               E,S: Scalar;
            begin
               S := ScalZero;
               for K in reverse D+1 .. D1 loop
                  E := ScalZero;
                  for I in IMax(0,K-P1.D) .. IMin(K,P2.D) loop
                     E := E+P2.C(I)*P1.C(K-I);
                  end loop;
                  S := Scal(R)*S+E;
               end loop;
               P3.C(D) := P3.C(D)+Scal(R*Err(S));
            end;
         end if;
         for K in reverse 0 .. D2 loop                     ---- add l.o. terms
            for I in IMax(0,K-P1.D) .. IMin(K,P2.D) loop
               P3.C(K) := P3.C(K)+P2.C(I)*P1.C(K-I);
            end loop;
         end loop;
      end if;
   end MultAdd;


   procedure MultSub(R: in Radius; P1,P2: in Taylor; P3: in out Taylor; D: Degree := Size) is
      D1: Natural;
      D2: Degree;
   begin
      if abs(P3.D)>1000 then New_Line; Put("Wrong P3.D!!!!"); end if;
      if abs(P2.D)>1000 then New_Line; Put("Wrong P2.D!!!!"); end if;
      if abs(P1.D)>1000 then New_Line; Put("Wrong P1.D!!!!"); end if;
      DegConvert(R,D,P3);                                 ---- now P3.D <= D
      if P1.D >= 0 and P2.D >= 0 then
         D1 := P1.D+P2.D;
         D2 := IMin(D1,D);
         if P3.D<D2 then                                   ---- need to fill in 0s
            for I in P3.D+1 .. D2 loop
               P3.C(I) := ScalZero;
            end loop;
            P3.D := D2;
         end if;
         if not Trunc and then D1>D then                   ---- collect h.o. terms
            declare
               E,S: Scalar;
            begin
               S := ScalZero;
               for K in reverse D+1 .. D1 loop
                  E := ScalZero;
                  for I in IMax(0,K-P1.D) .. IMin(K,P2.D) loop
                     E := E+P2.C(I)*P1.C(K-I);
                  end loop;
                  S := Scal(R)*S+E;
               end loop;
               P3.C(D) := P3.C(D)+Scal(R*Err(S));
            end;
         end if;
         for K in reverse 0 .. D2 loop                     ---- add l.o. terms
            for I in IMax(0,K-P1.D) .. IMin(K,P2.D) loop
               P3.C(K) := P3.C(K)-P2.C(I)*P1.C(K-I);
            end loop;
         end loop;
      end if;
   end MultSub;


   procedure Quot(P1,P2: in Taylor; P3: out Taylor; D: in Degree := Size) is
      R,S: Scalar;
   begin
      if not Trunc then
         Message("Taylor_Ops.Quot",Numeric_Only);
      end if;
      S := P2.C(0);
      if P2.D < 0 or else S = ScalZero then
         Message("Taylor_Ops.Quot",Division_By_Zero);
         return;
      end if;
      if D < 0 or else P1.D < 0 then
         P3.D := -1;
      else
         P3.D := D;
         P3.C(0) := P1.C(0)/S;
         for K in 1 .. D loop
            if K <= P1.D then
               R := P1.C(K);
            else
               R := ScalZero;
            end if;
            for I in 1 .. IMin(K,P2.D) loop
               R := R-P2.C(I)*P3.C(K-I);
            end loop;
            P3.C(K) := R/S;
         end loop;
      end if;
   end Quot;


   function Quot(P1,P2: Taylor; D: in Degree := Size) return Taylor is
      Pt: Taylor;
   begin
      Quot(P1,P2,Pt,D);
      return Pt;
   end Quot;


   function "/"(P1,P2: Taylor) return Taylor is
      Pt: Taylor;
   begin
      Quot(P1,P2,Pt);
      return Pt;
   end "/";


   procedure Comp(P1,P2: in Taylor; P3: out Taylor) is
      R3: Radius;
   begin
      if not Trunc then
         Message("Taylor_Ops.Comp",Numeric_Only);
      end if;
      if P1.D <= 0 then
         Copy(P1,P3);
      else
         R3 := Radius(1);                            ------ arbitrary
         Prod(P1.C(P1.D),P2,P3);
         P3.C(0) := P3.C(0)+P1.C(P1.D-1);
         for I in reverse 0 .. P1.D-2 loop
            Mult(R3,P2,P3);
            P3.C(0) := P3.C(0)+P1.C(I);
         end loop;
      end if;
   end Comp;


   function Comp(P1,P2: Taylor) return Taylor is
      Pt: Taylor;
   begin
      Comp(P1,P2,Pt);
      return Pt;
   end Comp;


   procedure InvComp(R,Err: in Radius; P1,P2: in Taylor; P3: in out Taylor) is
      --    iterate p3 -> p3-[p2(p3)-p1]/p2'(p3)
      P,Pd,Pd2,Pd3: Taylor;
   begin
      if not Trunc then
         Message("Taylor_Ops.InvComp",Numeric_Only);
      end if;
      if P3.D < 0 then
         P3.D := 0;
         P3.C(0) := ScalZero;
         FindRoot(Err,P2,Val(R,ScalZero,P1),P3.C(0));
      end if;
      Der(R,R,P2,Pd2);
      for I in 1 .. 100 loop
         Comp(P2,P3,P);
         Sub(P1,P);
         Comp(Pd2,P3,Pd);
         Quot(P,Pd,Pd3);
         Sub(Pd3,P3);
         if Sup(Norm(R,Pd3)) <= Err then
            return;
         end if;
      end loop;
      Message("Taylor_Ops.InvComp",Giving_Up);
   end InvComp;


   function InvComp(R,Err: Radius; P1,P2: Taylor) return Taylor is
      Pt: Taylor;
   begin
      Pt.D := -1;
      InvComp(R,Err,P1,P2,Pt);
      return Pt;
   end InvComp;


   function DerVal(R: Radius; Z: Scalar; P: Taylor) return Scalar is
      S: Scalar;
   begin
      if not Trunc then
         Message("Taylor_Ops.DerVal",Numeric_Only);
      end if;
      S := ScalZero;
      for I in reverse 1 .. P.D loop
         S := Z*S+Scal(I)*P.C(I);
      end loop;
      return S;
   end DerVal;


   procedure MinDeg(P1: in Taylor; P2: in out Taylor) is
   begin
      if not Trunc then
         Message("Taylor_Ops.MinDeg",Numeric_Only);
      end if;
      P2.D := Imin(P1.D,P2.D);
   end MinDeg;


   procedure CutDeg(R: in Radius; Cut: in Rep; P: in out Taylor) is
      C: Rep;
   begin
      if P.D>=0 then
         if not Trunc then
            Message("Taylor_Ops.CutDeg",Numeric_Only);
         end if;
         C := Cut*Exp(-Rep(P.D)*Log(R));
         while P.D>=0 and then Center(P.C(P.D))<=C loop
            C := C*R;
            P.D := P.D-1;
         end loop;
      end if;
   end CutDeg;


   function NumIprod(R: Radius; P1,P2: Taylor) return Scalar is
      S: Scalar;
   begin
      if not Trunc then
         Message("Taylor_Ops.NumIprod",Numeric_Only);
      end if;
      S := ScalZero;
      for K in reverse 0 .. Imin(P1.D,P2.D) loop
         S :=Scal(R)*Scal(R)*S+P1.C(K)*P2.C(K);
      end loop;
      return S;
   end NumIprod;


   function NumIprod(P1,P2: Taylor) return Scalar is
      S: Scalar;
   begin
      if not Trunc then
         Message("Taylor_Ops.NumIprod",Numeric_Only);
      end if;
      S := ScalZero;
      for K in reverse 0 .. Imin(P1.D,P2.D) loop
         S := S+P1.C(K)*P2.C(K);
      end loop;
      return S;
   end NumIprod;


   procedure FindRoot(Err: in Radius; P: in Taylor; S: in Scalar; R: in out Scalar) is
      --  iterate r -> r+[s-p(r)]/p'(r)
      R1: Radius;
      Dr: Scalar;
   begin
      if not Trunc then
         Message("Taylor_Ops.FindRoot",Numeric_Only);
      end if;
      R1 := Radius(1);                              ------ arbitrary
      for I in 1 .. 1000 loop
         Dr := (S-Val(R1,R,P))/DerVal(R1,R,P);
         R := R+Dr;
         if SupAbs(Dr) <= Err then
            return;
         end if;
      end loop;
      Message("Taylor_Ops.FindRoot",Giving_Up);
   end FindRoot;


   function Exp1 return Taylor is
      P: Taylor;
   begin
      if not Trunc then
         Message("Taylor_Ops.Exp1",Numeric_Only);
      end if;
      P.D := Size;
      P.C(0) := ScalZero;
      P.C(1) := ScalOne;
      for I in 2 .. Size loop
         P.C(I) := P.C(I-1)/Scal(I);
      end loop;
      return P;
   end Exp1;


   function Ln1 return Taylor is
      S: Scalar;
      P: Taylor;
   begin
      if not Trunc then
         Message("Taylor_Ops.Ln1",Numeric_Only);
      end if;
      P.D := Size;
      P.C(0) := ScalZero;
      S := ScalOne;
      for I in 1 .. Size loop
         P.C(I) := S/Scal(I);
         S := -S;
      end loop;
      return P;
   end Ln1;


   procedure NextRandom(I: in out Long_Integer) is
      Hi,Lo: Long_Integer;
   begin
      Hi := I/127773;
      Lo := I-127773*Hi;
      I := 16807*Lo-2836*Hi;
      if I < 0 then
         I := I+2147483647;
      end if;
   end NextRandom;


   procedure Random(D: in Degree; Seed: in out Long_Integer; R: in Radius; P: out Taylor) is
      Rp: Rep;
   begin
      P.D := D;
      Rp := One;
      for K in 0 .. D loop
         P.C(K) := Scal((One-(Two*Rep(Seed))/Rep(2147483647))*Rp);
         Rp := Rp/R;
         NextRandom(Seed);
      end loop;
   end Random;


   procedure RandomOdd(D: in Degree; Seed: in out Long_Integer; R: in Radius; P: out Taylor) is
      Rp: Rep;
   begin
      P.D := D;
      Rp := One;
      for K in 0 .. D loop
         if Float(0.5*Rep(K))=Float(Integer(0.5*Rep(K))) then
            P.C(K) := Scal(Zero);
         else
            P.C(K) := Scal((One-(Two*Rep(Seed))/Rep(2147483647))*Rp);
         end if;
         Rp := Rp/R;
         NextRandom(Seed);
      end loop;
   end RandomOdd;


   procedure RandomEven(D: in Degree; Seed: in out Long_Integer; R: in Radius; P: out Taylor) is
      Rp: Rep;
   begin
      P.D := D;
      Rp := One;
      for K in 0 .. D loop
         if Float(0.5*Rep(K))=Float(Integer(0.5*Rep(K))) then
            P.C(K) := Scal((One-(Two*Rep(Seed))/Rep(2147483647))*Rp);
         else
            P.C(K) := Scal(Zero);
         end if;
         Rp := Rp/R;
         NextRandom(Seed);
      end loop;
   end RandomEven;

end Taylor_Ops;

